package net.pricefx.logic.core;

import com.fasterxml.jackson.annotation.JsonInclude;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.MapperFeature;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.SerializationFeature;
import com.fasterxml.jackson.databind.introspect.Annotated;
import com.fasterxml.jackson.databind.introspect.JacksonAnnotationIntrospector;
import com.fasterxml.jackson.databind.ser.impl.SimpleBeanPropertyFilter;
import com.fasterxml.jackson.databind.ser.impl.SimpleFilterProvider;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.HashSet;
import java.util.Set;

import static net.pricefx.common.api.PriceFxInterface.FIELD_CREATEDATE;
import static net.pricefx.common.api.PriceFxInterface.FIELD_FE_FORMULAEXPRESSION;
import static net.pricefx.common.api.PriceFxInterface.FIELD_LASTUPDATEBYNAME;
import static net.pricefx.common.api.PriceFxInterface.FIELD_LASTUPDATEDATE;
import static net.pricefx.common.api.PriceFxInterface.FIELD_PL_LAST_UPDATE_BY;
import static net.pricefx.common.api.PriceFxInterface.FIELD_TYPEDID;
import static net.pricefx.common.api.PriceFxInterface.FIELD_VERSION;

/**
 * Various helper methods.
 */
public class Utils {

    private Utils() {
    }

    private static final Set<String> IGNORED_ATTRIBUTES = new HashSet<>();

    static {
        IGNORED_ATTRIBUTES.add(FIELD_FE_FORMULAEXPRESSION);
        IGNORED_ATTRIBUTES.add(FIELD_LASTUPDATEDATE);
        IGNORED_ATTRIBUTES.add(FIELD_PL_LAST_UPDATE_BY);
        IGNORED_ATTRIBUTES.add(FIELD_LASTUPDATEBYNAME);
        IGNORED_ATTRIBUTES.add(FIELD_TYPEDID);
        IGNORED_ATTRIBUTES.add(FIELD_VERSION);
        IGNORED_ATTRIBUTES.add(FIELD_CREATEDATE);
        IGNORED_ATTRIBUTES.add("createdBy");
        IGNORED_ATTRIBUTES.add("errors");
        IGNORED_ATTRIBUTES.add("diskFolder");
    }

    public static ObjectMapper getObjectMapper() {
        return new ObjectMapper()
                .configure(SerializationFeature.ORDER_MAP_ENTRIES_BY_KEYS, true)
                .configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false)
                .configure(MapperFeature.SORT_PROPERTIES_ALPHABETICALLY, true)
                .configure(SerializationFeature.INDENT_OUTPUT, true)
                .setSerializationInclusion(JsonInclude.Include.NON_NULL);
    }

    public static ObjectMapper getObjectMapperWithEmptyMaps() {
        return new ObjectMapper()
                .configure(SerializationFeature.ORDER_MAP_ENTRIES_BY_KEYS, true)
                .configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false)
                .configure(MapperFeature.SORT_PROPERTIES_ALPHABETICALLY, true)
                .configure(SerializationFeature.INDENT_OUTPUT, true);
    }

    public static ObjectMapper getObjectMapperUnsorted() {
        return new ObjectMapper()
                .configure(SerializationFeature.ORDER_MAP_ENTRIES_BY_KEYS, false)
                .configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false)
                .configure(MapperFeature.SORT_PROPERTIES_ALPHABETICALLY, true)
                .configure(SerializationFeature.INDENT_OUTPUT, true)
                .setSerializationInclusion(JsonInclude.Include.NON_NULL);
    }

    public static String serializeToJSONAsString(Object object) throws JsonProcessingException {
        ObjectMapper objectMapper = getObjectMapperWithEmptyMaps();
        SimpleBeanPropertyFilter filter = SimpleBeanPropertyFilter.serializeAllExcept(IGNORED_ATTRIBUTES);
        SimpleFilterProvider filterProvider = new SimpleFilterProvider()
                .addFilter(FilterOutIgnoredFieldsInstrospector.FILTER_ID, filter);

        objectMapper = objectMapper.setAnnotationIntrospector(new FilterOutIgnoredFieldsInstrospector());
        return objectMapper.writer(filterProvider).writeValueAsString(object);
    }

    public static byte[] serializeToJSONAsBytes(Object object) throws IOException {
        return serializeToJSONAsString(object).getBytes(StandardCharsets.UTF_8);
    }

    public static long convertTypedIdToId(String typedId) {
        String[] pieces = typedId.split("\\.");
        return Long.parseLong(pieces[0]);
    }

    public static String convertTypedIdToTypeCode(String typedId) {
        return typedId.substring(typedId.indexOf('.') + 1);
    }

    private static class FilterOutIgnoredFieldsInstrospector extends JacksonAnnotationIntrospector {

        public static final String FILTER_ID = "FILTER_OUT_IGNORED_FIELDS";

        @Override
        public Object findFilterId(Annotated ac) {
            return FILTER_ID;
        }
    }

}
