package net.pricefx.logic.core.model;

import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.databind.node.ObjectNode;
import net.pricefx.common.api.IFormulaNature;
import net.pricefx.common.api.PriceFxInterface;
import net.pricefx.pckg.transform.TransformCalculationLogic;
import net.pricefx.pckg.transform.TransformWorkflowFormula;
import net.pricefx.pckg.transform.descriptor.TypeDescriptor;

import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import static net.pricefx.common.api.FormulaNature.CO_HEADER;
import static net.pricefx.common.api.FormulaNature.CONTRACT_HEADER;
import static net.pricefx.common.api.FormulaNature.QUOTE_HEADER;
import static net.pricefx.common.api.FormulaNature.REBATE_AGREEMENT_HEADER;
import static net.pricefx.common.api.PriceFxInterface.FIELD_FE_ELEMENTS;
import static net.pricefx.common.api.PriceFxInterface.FIELD_FE_FORMULANATURE;
import static net.pricefx.common.api.PriceFxInterface.FIELD_LABEL;
import static net.pricefx.common.api.PriceFxInterface.FIELD_SIMULATIONSET;
import static net.pricefx.common.api.PriceFxInterface.FIELD_STATUS;
import static net.pricefx.common.api.PriceFxInterface.FIELD_TYPEDID;
import static net.pricefx.common.api.PriceFxInterface.FIELD_UNIQUENAME;
import static net.pricefx.common.api.PriceFxInterface.FIELD_VALIDAFTER;
import static net.pricefx.common.api.PriceFxInterface.FIELD_VERSION;


public abstract class AbstractFormula<E extends AbstractFormulaElement> extends ObjectHolder implements Iterable<AbstractFormulaElement> {

	public static final String LOGIC_NAME = "StudioEvaluator";
	public static final String ELEMENT_DEBUG = "result";

	protected AbstractFormula() {
		super();
		getOrCreateElements();
	}

	protected AbstractFormula(ObjectNode logic) {
		super(logic);
		getOrCreateElements();
	}

	public static AbstractFormula createLogic(ObjectNode node, TypeDescriptor objectType, Path modulePath) {
		AbstractFormula logic;

		if (objectType == TransformCalculationLogic.DESCRIPTOR) {
			logic = new PricingFormula(node);
		} else if (objectType == TransformWorkflowFormula.DESCRIPTOR) {
			logic = new WorkflowFormula(node);
		} else {
			throw new IllegalArgumentException("Cannot create logic for " + objectType.getName());
		}

		logic.setDirPathBasedOnFsId(modulePath);

		return logic;
	}

	public TypeDescriptor getTypeDescriptor() {
		if (this instanceof PricingFormula) {
			return TransformCalculationLogic.DESCRIPTOR;
		} else if (this instanceof WorkflowFormula) {
			return TransformCalculationLogic.DESCRIPTOR;
		} else {
			throw new IllegalArgumentException("Cannot get TypeDescriptor for " + this.getUniqueName());
		}
	}

	public static AbstractFormula createLogic(TypeDescriptor objectType) {
		if (objectType == TransformCalculationLogic.DESCRIPTOR) {
			return new PricingFormula();
		} else if (objectType == TransformWorkflowFormula.DESCRIPTOR) {
			return new WorkflowFormula();
		} else {
			throw new IllegalArgumentException("Cannot create logic for " + objectType.getName());
		}
	}

	public static AbstractFormula createVirtualLogic(TypeDescriptor objectType) {
		AbstractFormula virtualLogic = createLogic(objectType);
		virtualLogic.setUniqueName(LOGIC_NAME);

		AbstractFormulaElement element = virtualLogic.createElement();
		element.setElementName(ELEMENT_DEBUG);
		element.setCombinationType(ElementCombinationType.FUNCTION.name());
		virtualLogic.addElement(element);

		return virtualLogic;
	}

	public AbstractFormulaElement getElementByName(String elementName) {
		if (elementName == null) {
			return null;
		}

		Iterator<AbstractFormulaElement> it = getElementsIterator();
		while (it.hasNext()) {
			AbstractFormulaElement element = it.next();

			if (elementName.equals(element.getElementName())) {
				return element;
			}
		}
		return null;
	}

	public int getElementIndexByName(String elementName) {
		if (elementName == null) {
			return -1;
		}

		Iterator<AbstractFormulaElement> it = getElementsIterator();
		int i = 0;
		while (it.hasNext()) {
			AbstractFormulaElement element = it.next();

			if (elementName.equals(element.getElementName())) {
				return i;
			}
			i++;
		}
		return -1;
	}

	@Override
	public Iterator<AbstractFormulaElement> iterator() {
		return new ElementIterator();
	}


    public class ElementIterator implements Iterator<AbstractFormulaElement> {
		private final Iterator<Map<String, Object>> it = (getElements() == null ? new ArrayList().iterator() : getElements().iterator());

		@Override
		public AbstractFormulaElement next() {
			return createElement(it.next());
		}

		@Override
		public boolean hasNext() {
			return it.hasNext();
		}

	}

	public Iterator<AbstractFormulaElement> getElementsIterator() {
		return iterator();
	}

	public List getOrCreateElements() {
		Object val = holder.get(FIELD_FE_ELEMENTS);
		if (val instanceof List) {
			return (List) val;
		}
		List elements = new ArrayList<>();
		this.holder.put(FIELD_FE_ELEMENTS, elements);
		return elements;
	}

	public void addElement(AbstractFormulaElement element) {
		getOrCreateElements().add(element.getHolder());
	}


	public abstract AbstractFormulaElement createElement();

	public abstract AbstractFormulaElement createElement(Map<String, Object> node);

	public abstract boolean isPricingLogic();

	public abstract boolean isWorkflowLogic();

	public boolean isHeaderLogic() {
		IFormulaNature nature = getFormulaNatureObject();

		return QUOTE_HEADER.equals(nature)
				|| CONTRACT_HEADER.equals(nature)
				|| REBATE_AGREEMENT_HEADER.equals(nature)
				|| CO_HEADER.equals(nature);
	}


	public String getUniqueName() {
		return getTextValue(FIELD_UNIQUENAME);
	}

	public void setUniqueName(String uniqueName) {
		setTextValue(FIELD_UNIQUENAME, uniqueName);
	}

	public String getLabel() {
		return getTextValue(FIELD_LABEL);
	}

	public void setLabel(String label) {
		setTextValue(FIELD_LABEL, label);
	}

	public List<Map<String, Object>> getElements() {
		return getArrayValue(FIELD_FE_ELEMENTS);
	}

	public String getValidAfter() {
		return getTextValue(FIELD_VALIDAFTER);
	}

	public void setValidAfter(String validAfter) {
		setTextValue(FIELD_VALIDAFTER, validAfter);
	}

	public String getStatus() {
		return getTextValue(FIELD_STATUS);
	}

	public void setStatus(String status) {
		setTextValue(FIELD_STATUS, status);
	}

	public Long getVersion() {
		return getLongValue(FIELD_VERSION);
	}

	public void setVersion(Long version) {
		setLongValue(FIELD_VERSION, version);
	}

	public String getTypedId() {
		return getTextValue(FIELD_TYPEDID);
	}

	public void setTypedId(String typedId) {
		setTextValue(FIELD_TYPEDID, typedId);
	}

	public String getSimulationSet() {
		return getTextValue(FIELD_SIMULATIONSET);
	}

	public void setSimulationSet(String simulationSet) {
		setTextValue(FIELD_SIMULATIONSET, simulationSet);
	}

	public String getFormulaType() {
		return getTextValue("formulaType");
	}

	public void setFormulaType(String formulaType) {
		setTextValue("formulaType", formulaType);
	}

	public String getFormulaNature() {
		return getTextValue(FIELD_FE_FORMULANATURE);
	}

	public String getUserGroupEdit(){
		return getTextValue(PriceFxInterface.FIELD_USERGROUPEDIT);
	}
	public void setUserGroupEdit(String userGroupEdit){
		setTextValue(PriceFxInterface.FIELD_USERGROUPEDIT, userGroupEdit);
	}

	public String getUserGroupViewDetails(){
		return getTextValue(PriceFxInterface.FIELD_USERGROUPVIEWDETAILS);
	}
	public void setUserGroupViewDetails(String userGroupViewDetails){
		setTextValue(PriceFxInterface.FIELD_USERGROUPVIEWDETAILS, userGroupViewDetails);
	}

	public void setFormulaNature(String formulaNature) {
		setTextValue(FIELD_FE_FORMULANATURE, formulaNature);
	}

	@JsonIgnore
	public abstract IFormulaNature getFormulaNatureObject();

	@JsonIgnore
	public abstract TypeDescriptor getDescriptor();
}
